#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.securitypolicy.policy
"""
import unittest
import doctest

from zope.interface import implements, directlyProvides, Interface
from zope.component import adapts
from zope.component import provideAdapter, provideUtility
from zope.app.testing import setup
from zope.traversing.interfaces import IContainmentRoot

from schooltool.securitypolicy.crowds import Crowd
from schooltool.securitypolicy.interfaces import ICrowd


class IObj(Interface):
    pass


class Obj(object):
    implements(IObj)

class AnotherObj(object):
    pass


class ParticipationStub(object):
    interaction = None
    principal = 'guest'


def test_SchoolToolSecurityPolicy_checkPermission():
    """Tests for SchoolToolSecurityPolicy.

        >>> from schooltool.securitypolicy.crowds import CrowdsUtility
        >>> from schooltool.securitypolicy.interfaces import ICrowdsUtility
        >>> cru = CrowdsUtility()
        >>> provideUtility(cru, ICrowdsUtility)

    Let's construct a security policy.

        >>> from schooltool.securitypolicy import policy
        >>> participation = ParticipationStub()
        >>> sp = policy.SchoolToolSecurityPolicy(participation)

        >>> byadaptation_returns = False
        >>> def checkByAdaptationStub(permission, obj):
        ...     print 'checkByAdaptation' + str((permission, obj))
        ...     return byadaptation_returns
        >>> sp.checkByAdaptation = checkByAdaptationStub

        >>> checkcrowds_returns = False
        >>> def checkCrowdsStub(crowdclasses, obj):
        ...     print 'checkCrowds' + str((crowdclasses, obj))
        ...     return checkcrowds_returns
        >>> sp.checkCrowds = checkCrowdsStub

    First, we test the path where no crowds accept the principal:

        >>> obj = Obj()

        >>> sp.checkPermission('perm', obj)
        checkByAdaptation('perm', <...Obj object ...>)
        False

    Interface-independent permissions
    ---------------------------------

        >>> checkcrowds_returns = True
        >>> cru.factories = {'the crowd': 'crowd factory'}
        >>> cru.crowds[('perm', None)] = ['the crowd']
        >>> sp.checkPermission('perm', obj)
        checkCrowds(['crowd factory'], <...Obj object ...>)
        True

    Another case: there is a registered crowd for the permission, but checkCrowds
    fails:

        >>> checkcrowds_returns = False

        >>> sp.checkPermission('perm', obj)
        checkCrowds(['crowd factory'], <...Obj object ...>)
        False

    """


def test_SchoolToolSecurityPolicy_checkByAdaptation():
    """Tests for SchoolToolSecurityPolicy.checkByAdaptation.

        >>> from schooltool.securitypolicy.crowds import CrowdsUtility
        >>> from schooltool.securitypolicy.interfaces import ICrowdsUtility
        >>> cru = CrowdsUtility()
        >>> provideUtility(cru, ICrowdsUtility)

    Let's construct a security policy.

        >>> from schooltool.securitypolicy import policy
        >>> participation = ParticipationStub()
        >>> sp = policy.SchoolToolSecurityPolicy(participation)

        >>> obj = Obj()

    First, check the straightforward case

        >>> contains_returns = False
        >>> class ObjCrowd(Crowd):
        ...     adapts(IObj)
        ...     def contains(self, principal):
        ...         return contains_returns
        >>> provideAdapter(ObjCrowd, (IObj,), ICrowd, 'perm')

        >>> sp.checkByAdaptation('perm', obj)
        False

        >>> contains_returns = True
        >>> sp.checkByAdaptation('perm', obj)
        True

    If no crowd is found for a given object, its parents are checked instead:

        >>> obj2 = AnotherObj()
        >>> obj2.__parent__ = obj
        >>> obj3 = AnotherObj()
        >>> obj3.__parent__ = obj2

        >>> sp.checkPermission('perm', obj3)
        True

    The code won't climb up beyond IContainmentRoot:

        >>> obj2.__parent__ = None
        >>> directlyProvides(obj2, IContainmentRoot)
        >>> sp.checkPermission('perm', obj3)
        Traceback (most recent call last):
            ...
        AssertionError: ('no crowd found for', None, 'perm')

    """


def test_SchoolToolSecurityPolicy_checkCrowds():
    """Tests for SchoolToolSecurityPolicy.checkCrowds.

    Check crowds just check whether any principal is contained in at
    least one of the crowds constructed from a crowdcls list passed to
    it:

        >>> class NegativeCrowdStub(object):
        ...     def __init__(self, context):
        ...         pass
        ...     def contains(self, principal):
        ...         return False
        >>> class CrowdStub(NegativeCrowdStub):
        ...     def contains(self, principal):
        ...         return principal == 'john'

        >>> class ParticipationStub(object):
        ...     interaction = None
        ...     def __init__(self, principal):
        ...         self.principal = principal

        >>> from schooltool.securitypolicy import policy
        >>> sp = policy.SchoolToolSecurityPolicy(ParticipationStub('john'),
        ...                                      ParticipationStub('pete'))

        >>> sp.checkCrowds([NegativeCrowdStub, NegativeCrowdStub], object())
        False

        >>> sp.checkCrowds([NegativeCrowdStub, CrowdStub], object())
        True

    If john is not logged in, we'd still get False:

        >>> sp = policy.SchoolToolSecurityPolicy(ParticipationStub('pete'))
        >>> sp.checkCrowds([NegativeCrowdStub, CrowdStub], object())
        False

    """


def setUp(test=None):
    setup.placelessSetUp()

def tearDown(test=None):
    setup.placelessTearDown()


def test_suite():
    return unittest.TestSuite([
            doctest.DocTestSuite(optionflags=doctest.ELLIPSIS,
                                 setUp=setUp, tearDown=tearDown),
           ])



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from zope.cachedescriptors.property import Lazy
from zope.interface import directlyProvides
from zope.traversing.browser.absoluteurl import absoluteURL

import zc.table.column
from zc.table.interfaces import ISortableColumn

from schooltool import table
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.skin import flourish
from schooltool.person.interfaces import IPerson
from schooltool.task.interfaces import IRemoteTask
from schooltool.task.state import TaskReadState

from schooltool.common import SchoolToolMessage as _


class TaskContainerView(flourish.page.Page):

    @property
    def tasks(self):
        return sorted(self.context.values(), key=lambda t: (str(t.scheduled), t.task_id))


class TaskStatusView(flourish.page.Page):

    @property
    def status(self):
        return TaskReadState(self.task_id)

    @property
    def persistent_failed(self):
        task = self.persistent_task
        if task is None:
            return False
        return task.permanent_traceback is not None

    @property
    def persistent_finished(self):
        task = self.persistent_task
        if task is None:
            return False
        return task.permanent_result is not None

    @property
    def persistent_task(self):
        return self.context

    @property
    def task_id(self):
        return self.context.task_id


class TaskContainerLinkViewlet(flourish.page.LinkViewlet):

    def url(self):
        app = ISchoolToolApplication(None)
        base_url = absoluteURL(app, self.request)
        return '%s/%s' % (base_url, 'schooltool.tasks')


def task_state_formatter(task, item, formatter):
    result = []

    cls = ""
    if task.failed:
        cls = "error"
    elif task.succeeded or task.working:
        cls = "success"
    else:
        cls = "info"
    internal_state = task.internal_state or ''
    if internal_state:
        result.append(
            '<p%s>%s</p>' % (
                (' class="%s"') % cls if cls else '', internal_state))
    if task.permanent_traceback is not None:
        result.append('<p class="error">TRACEBACK</p>')
    if task.permanent_result is not None:
        result.append('<p class="success">PERSISTED RESULT</p>')
    return '\n'.join(result)


def task_id_cell_formatter(value, item, formatter):
    url = absoluteURL(item, formatter.request)
    result = '''
      <p>
        <a href=%s>%s</a>
      </p>
    ''' % (url, value)
    return result


def task_id_debug_cell_formatter(value, item, formatter):
    result = task_id_cell_formatter(value, item, formatter)
    app_url = absoluteURL(ISchoolToolApplication(None), formatter.request)
    result += '''
      <p>
        <a href="%s/schooltool.task_results/%s">%s</a>
      </p>
      ''' % (app_url, value, _('JSON result'))
    return result


class TaskTable(table.ajax.IndexedTable):

    no_default_url_cell_formatter = True
    task_id_formatter = lambda self, *a: task_id_cell_formatter(*a)

    def sortOn(self):
        return (('scheduled', True),)

    def columns(self):
        task_id = table.column.IndexedGetterColumn(
            index='task_id',
            name='task_id',
            cell_formatter=self.task_id_formatter,
            title=_(u'Task ID'),
            getter=lambda i, f: i.task_id,
            subsort=True)
        signature = table.table.GetterColumn(
            name='signature',
            title=_(u'Signature'),
            getter=lambda i, f: i.signature,
            subsort=True)
        internal_state = table.column.IndexedGetterColumn(
            index='internal_state',
            name='internal_state',
            title=_(u'Internal state'),
            getter=lambda i, f: i,
            cell_formatter=task_state_formatter,
            subsort=True)
        directlyProvides(internal_state, ISortableColumn)
        scheduled = table.column.IndexedGetterColumn(
            index='scheduled',
            name='scheduled',
            title=_(u'Scheduled'),
            getter=lambda i, f: i.scheduled,
            cell_formatter=table.table.datetime_formatter,
            subsort=True)
        directlyProvides(scheduled, ISortableColumn)
        return [task_id, signature, internal_state, scheduled]

    def setUp(self, *args, **kw):
        super(TaskTable, self).setUp(*args, **kw)
        self.css_classes['table'] = 'data schooltool-tasks-status'


class TaskTableDevmode(TaskTable):

    task_id_formatter = lambda self, *a: task_id_debug_cell_formatter(*a)


class MessageColumn(zc.table.column.GetterColumn):

    def renderCell(self, item, formatter):
        content = flourish.content.queryContentProvider(
            item, formatter.request, getattr(formatter, 'view', None),
            'short')
        if content is None:
            return ''
        result = content()
        return result


class MessageDialog(flourish.form.Dialog):

    template = flourish.templates.File('templates/task_dialog.pt')
    refresh_delay = 10000

    @Lazy
    def form_id(self):
        return flourish.page.obj_random_html_id(self)


class TaskProgressDialog(flourish.form.Dialog):

    template = flourish.templates.File('templates/f_task_progress.pt')

    @property
    def main_recipient(self):
        person = IPerson(self.request, None)
        if self.context.recipients is None:
            return None
        recipients = sorted(self.context.recipients, key=lambda r: r.__name__)
        if person in recipients:
            return person
        for recipient in recipients:
            if flourish.canView(recipient):
                return recipient
        return None

    @property
    def task_id(self):
        return self.task and self.task.__name__

    @property
    def progress_id(self):
        return flourish.page.sanitize_id('progress-%s' % (self.task_id or self.__name__))

    @Lazy
    def task(self):
        sender = self.context.sender
        if IRemoteTask.providedBy(sender):
            return sender
        return None

    @property
    def failed(self):
        return self.task and self.task.failed

    @property
    def completed(self):
        return (self.task and (self.task.succeeded or
                               self.task.finished and not self.task.failed))

    @property
    def pending(self):
        return self.task and not self.task.finished

    @property
    def should_reload(self):
        sender = self.context.sender
        if IRemoteTask.providedBy(sender):
            return not sender.finished
        return False



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Relationship maintenance.

Zope 3 content management operations (Delete, Cut, Copy, Paste) do not work
well with relationships unless you register event subscribers that perform
the necessary relationship maintenance.

This module contains event subscribers for IObjectRemovedEvent (and, soon,
others) that update relationships on these changes.
"""

from zope.container.contained import getProxiedObject
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.lifecycleevent.interfaces import IObjectCopiedEvent

from schooltool.relationship.interfaces import IRelationshipLinks
from schooltool.relationship.relationship import unrelateAll


def unrelateOnDeletion(event):
    """Remove all relationships when an object is deleted."""
    if not IObjectRemovedEvent.providedBy(event):
        return
    linkset = IRelationshipLinks(event.object, None)
    if linkset is not None:
        # event.object may be a ContainedProxy
        unrelateAll(getProxiedObject(event.object))


def unrelateOnCopy(event):
    """Remove all relationships when an object is copied."""
    if not IObjectCopiedEvent.providedBy(event):
        return
    # event.object may be a ContainedProxy
    obj = getProxiedObject(event.object)
    linkset = IRelationshipLinks(obj, None)
    if linkset is not None:
        links_to_remove = []
        for link in linkset:
            other_linkset = IRelationshipLinks(link.target)
            try:
                other_linkset.find(link.role, obj, link.my_role, link.rel_type)
            except ValueError:
                # The corresponding other link was not copied, so we have a
                # degenerate one-sided relationship.  Let's remove it
                # altogether.  It would not difficult to have a different
                # function, cloneRelationshipsOnCopy, that would create
                # a corresponding link in other_linkset.
                links_to_remove.append(link)
        for link in links_to_remove:
            linkset.remove(link)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Course implementation
"""

from persistent import Persistent

from zope.interface import implementer
from zope.interface import implements
from zope.component import adapts
from zope.component import adapter
from zope.component import getUtility
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.intid.interfaces import IIntIds
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.container.contained import Contained
from zope.container.btree import BTreeContainer

from schooltool.term.interfaces import ITerm
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.relationship import RelationshipProperty
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.app import InitBase
from schooltool.app import relationships
from schooltool.app.app import Asset
from schooltool.course.interfaces import ICourse
from schooltool.course.interfaces import ICourseContainer
from schooltool.course import interfaces


class CourseContainerContainer(BTreeContainer):
    """Container of Courses."""

    implements(interfaces.ICourseContainerContainer,
                              IAttributeAnnotatable)


class CourseContainer(BTreeContainer):
    """Container of Courses."""

    implements(interfaces.ICourseContainer,
                              IAttributeAnnotatable)


@adapter(ISchoolToolApplication)
@implementer(ICourseContainer)
def getCourseContainerForApp(app):
    syc = ISchoolYearContainer(app)
    sy = syc.getActiveSchoolYear()
    if sy is not None:
        return ICourseContainer(sy)


@adapter(ISchoolYear)
@implementer(ICourseContainer)
def getCourseContainer(sy):
    int_ids = getUtility(IIntIds)
    sy_id = str(int_ids.getId(sy))
    app = ISchoolToolApplication(None)
    cc = app['schooltool.course.course'].get(sy_id, None)
    if cc is None:
        cc = app['schooltool.course.course'][sy_id] = CourseContainer()
    return cc


@adapter(ITerm)
@implementer(ICourseContainer)
def getCourseContainerForTerm(term):
    return ICourseContainer(ISchoolYear(term))


@adapter(ICourse)
@implementer(ICourseContainer)
def getCourseContainerForCourse(course):
    return course.__parent__


@adapter(ICourseContainer)
@implementer(ISchoolYear)
def getSchoolYearForCourseContainer(course_container):
    container_id = int(course_container.__name__)
    int_ids = getUtility(IIntIds)
    container = int_ids.getObject(container_id)
    return container


@adapter(ICourse)
@implementer(ISchoolYear)
def getSchoolYearForCourse(course):
    return ISchoolYear(course.__parent__)


class Course(Persistent, Contained, Asset):

    implements(interfaces.ICourseContained, IAttributeAnnotatable)

    sections = RelationshipProperty(relationships.URICourseSections,
                                    relationships.URICourse,
                                    relationships.URISectionOfCourse)

    course_id = None
    government_id = None
    credits = None

    def __init__(self, title=None, description=None):
        self.title = title
        self.description = description


class CourseInit(InitBase):

    def __call__(self):
        self.app['schooltool.course.course'] = CourseContainerContainer()


class RemoveCoursesWhenSchoolYearIsDeleted(ObjectEventAdapterSubscriber):
    adapts(IObjectRemovedEvent, ISchoolYear)

    def __call__(self):
        course_container = ICourseContainer(self.object)
        for course_id in list(course_container.keys()):
            del course_container[course_id]



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Upgrade SchoolTool to generation 30.

Groups are no longer allowed as members of a section, so add members of a group
directly to the section.

Also, courses gained several new attributes, so fill in their course_id.
"""
import transaction

from zope.app.generations.utility import findObjectsProviding
from zope.app.publication.zopepublication import ZopePublication

from schooltool.group.interfaces import IGroup
from schooltool.course.interfaces import ICourse
from schooltool.relationship.relationship import getRelatedObjects
from schooltool.relationship.relationship import relate, unrelate
from schooltool.app.membership import URIMember, URIGroup, URIMembership


def evolve(context):
    root = context.connection.root().get(ZopePublication.root_name, None)

    savepoint_counter = 0

    # Groups are no longer allowed as members of a section.
    groups = findObjectsProviding(root, IGroup)
    for group in groups:
        targets = getRelatedObjects(
            group, URIGroup, rel_type=URIMembership)
        group_members = getRelatedObjects(
            group, URIMember, rel_type=URIMembership)

        for target in targets:
            target_members = getRelatedObjects(
                target, URIMember, rel_type=URIMembership)
            for member in group_members:
                if member not in target_members:
                    relate(URIMembership,
                           (member, URIMember),
                           (target, URIGroup))
                savepoint_counter += 1
            unrelate(URIMembership,
                     (group, URIMember),
                     (target, URIGroup))
        if savepoint_counter % 2000 == 0:
            transaction.savepoint(optimistic=True)

    # Courses gained several new attributes, fill in their course_id
    courses = findObjectsProviding(root, ICourse)
    for course in courses:
        if getattr(course, 'course_id', None) is None:
            course.course_id = course.__name__



from zope.app.form.browser import PasswordWidget
from zope.app.form.browser.widget import renderElement
from zope.app.form.interfaces import ConversionError
from zope.i18n import translate

from schooltool.common import SchoolToolMessage as _

class PasswordConfirmationWidget(PasswordWidget):
    """Password Widget that uses two fields to confirm user input.
    """

    default = ''
    type = 'password'
    displayWidth = 15

    def __call__(self):
        password_tag = renderElement(
                self.tag,
                type=self.type,
                name=self.name,
                id=self.name,
                value=self.default,
                cssClass=self.cssClass,
                style=self.style,
                size=self.displayWidth,
                extra=self.extra)
        confirm_name = self.name + '.confirm'
        confirm_tag = renderElement(
                self.tag,
                type=self.type,
                name=confirm_name,
                id=confirm_name,
                value=self.default,
                cssClass=self.cssClass,
                style=self.style,
                size=self.displayWidth,
                extra=self.extra)

        confirm_label = translate(_("Confirm password"), context=self.request)
        return u"""<div><div>%s</div>
                     <br />
                     <div class="label">
                       <label for="%s" title="">%s</label>
                     </div>
                     <div>%s</div>
                   </div>""" % (
            password_tag, confirm_name, confirm_label, confirm_tag)

    def _toFieldValue(self, input):
        """Check whether the confirmation field value is identical to
        the password value.
        """
        formdata = self.request.form
        if input != formdata[self.name + '.confirm']:
            raise ConversionError(_(u"Supplied passwords are not identical"),
                                  '')
        return super(PasswordConfirmationWidget, self)._toFieldValue(input)

    def hasInput(self):
        """Check whether the field is represented in the form.
        """
        return self.name + ".confirm" in self.request.form or \
          super(PasswordConfirmationWidget, self).hasInput()

    def hidden(self):
        raise NotImplementedError(
            'Cannot get a hidden tag for a password field')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for common widgets.
"""
import unittest
import doctest

import zope.component
import zope.interface
from zope.interface.verify import verifyObject
from zope.publisher.browser import TestRequest

from z3c import form

from schooltool.schoolyear.ftesting import schoolyear_functional_layer
import schooltool.skin.widgets
from schooltool import skin
from schooltool.app.browser.testing import setUp, tearDown


def setUpAppAbsoluteURL():
    from schooltool.app.interfaces import ISchoolToolApplication
    from zope.traversing.browser.interfaces import IAbsoluteURL
    from zope.publisher.interfaces.http import IHTTPRequest
    class App(object):
        zope.interface.implements(ISchoolToolApplication)
    zope.component.provideAdapter(
        lambda ignored: App(),
        adapts=(None, ), provides=ISchoolToolApplication)
    zope.component.provideAdapter(
        lambda app, request: lambda: request.getApplicationURL(),
        adapts=(App, IHTTPRequest), provides=IAbsoluteURL)


class FormRequest(TestRequest):
    zope.interface.implements(skin.skin.ISchoolToolLayer)


def doctest_HTMLFragmentWidget():
    """Tests for HTMLFragmentWidget.

        >>> widget = skin.widgets.HTMLFragmentWidget()
        >>> verifyObject(skin.widgets.IHTMLFragmentWidget, widget)
        True

    """


def doctest_CkeditorZ3CFormWidget_CkeditorConfig():
    """Tests for CkeditorZ3CFormWidget and CkeditorConfig.

        >>> setUpAppAbsoluteURL()

    We'll need a field first.

        >>> from zope.html.field import HtmlFragment
        >>> schema_field = HtmlFragment(__name__='html', title=u"Fragment")

    Let's build a widget bound to the field.

        >>> zope.component.provideAdapter(skin.widgets.CkeditorFieldWidget)

        >>> request = FormRequest()
        >>> widget = zope.component.getMultiAdapter(
        ...     (schema_field, request), form.interfaces.IFieldWidget)

        >>> print widget
        <CkeditorZ3CFormWidget 'html'>

        >>> verifyObject(skin.widgets.ICkeditorWidget, widget)
        True

    Widget initially has no config set, so it cannot render the CKEditor
    javascript setup.

        >>> print widget.config
        None

        >>> widget.script
        Traceback (most recent call last):
        ...
        AttributeError: ...

    Config will be set during widget update, as a computed widget value.

        >>> zope.component.provideAdapter(
        ...     skin.widgets.Ckeditor_config, name='config')

        >>> value = zope.component.getMultiAdapter(
        ...     (widget.context, widget.request,
        ...      widget.form, widget.field, widget),
        ...     form.interfaces.IValue, name='config')

        >>> print value
        <ComputedValue <schooltool.skin.widgets.CkeditorConfig
          (430 x 300, u'schooltool' toolbar, u'/@@/editor_config.js')>>

        >>> verifyObject(skin.widgets.ICkeditorConfig, value.get())
        True

        >>> widget.update()

        >>> print widget.config
        <schooltool.skin.widgets.CkeditorConfig
          (430 x 300, u'schooltool' toolbar, u'/@@/editor_config.js')>

    Now we can render the javascript.

        >>> print widget.script
        <script type="text/javascript" language="JavaScript">
            var CKeditor... = new CKEDITOR.replace("html",
                {
                    height: 300,
                    width: 430,
                    customConfig: "http://127.0.0.1/@@/editor_config.js"
                }
            );
        </script>

    Let's set the widget value.

        >>> widget.update()

        >>> widget.value
        u''

        >>> request = FormRequest(
        ...     form={widget.name: '<strong>All hail hypnotoad!</strong>'})

        >>> widget = zope.component.getMultiAdapter(
        ...     (schema_field, request), form.interfaces.IFieldWidget)

        >>> widget.update()
        >>> print widget.value
        <strong>All hail hypnotoad!</strong>

    We also have a different configuration for add and edit forms.

        >>> zope.component.provideAdapter(
        ...     skin.widgets.Ckeditor_addform_config, name='config')
        >>> zope.component.provideAdapter(
        ...     skin.widgets.Ckeditor_editform_config, name='config')

        >>> widget.form = form.form.AddForm(None, request)
        >>> widget.update()
        >>> print widget.config
        <schooltool.skin.widgets.CkeditorConfig
          (306 x 200, u'schooltool' toolbar, u'/@@/editor_config.js')>

        >>> verifyObject(skin.widgets.ICkeditorConfig, widget.config)
        True

        >>> widget.form = form.form.EditForm(None, request)
        >>> widget.update()
        >>> print widget.config
        <schooltool.skin.widgets.CkeditorConfig
          (306 x 200, u'schooltool' toolbar, u'/@@/editor_config.js')>

    """


def doctest_CkeditorZ3CWidget_compatibility():
    r"""Tests for CkeditorZ3CFormWidget compatibility with zope.html.

        >>> setUpAppAbsoluteURL()

    We'll need a field first.

        >>> from zope.html.field import HtmlFragment
        >>> schema_field = HtmlFragment(__name__='html', title=u"Fragment")

    Let's build a widget bound to the field.

        >>> zope.component.provideAdapter(skin.widgets.CkeditorFieldWidget)

        >>> request = FormRequest()
        >>> widget = zope.component.getMultiAdapter(
        ...     (schema_field, request), form.interfaces.IFieldWidget)

        >>> verifyObject(skin.widgets.ICkeditorWidget, widget)
        True

    Specify the CkeditorConfig.

        >>> zope.component.provideAdapter(
        ...     skin.widgets.Ckeditor_config, name='config')

    Now we can render the javascript.

        >>> widget.update()
        >>> print widget.script
        <script type="text/javascript" language="JavaScript">
        ...
        </script>

    Let's now look at a fully rendered zope.html widget.

        >>> from zope.html.widget import CkeditorWidget
        >>> editor = CkeditorWidget(schema_field, request)

        >>> print editor()
        <textarea cols="60" id="field.html" name="field.html" rows="15" ></textarea>
        <script type="text/javascript" language="JavaScript">
        var CKeditor_html = new CKEDITOR.replace("field.html",
            {
                height: 400,
                customConfig : "/@@/zope_ckconfig.js",
            }
        );
        </script>

    We can also modify the configuration of the zope.html widget.

        >>> editor.editorWidth = 430
        >>> editor.editorHeight = 300
        >>> editor.toolbarConfiguration = "schooltool"
        >>> editor.configurationPath = '/@@/editor_config.js'

        >>> print editor()
        <textarea cols="60" id="field.html" name="field.html" rows="15" ></textarea>
        <script type="text/javascript" language="JavaScript">
        var CKeditor_html = new CKEDITOR.replace("field.html",
            {
                height: 300,
                customConfig : "/@@/editor_config.js",
            }
        );
        </script>

    You can notice zope.html widget uses relative paths for configuration;
    this breaks the widget if admins start fiddling with Apache's mod-rewrite
    (to put schooltool in http://example.com/schooltool for example).

    Our widget also uses a different mechanism to generate the CKeditor
    variable name.

        >>> import difflib
        >>> def print_diff(old, new):
        ...     stripped = lambda s: [l.strip() for l in s.splitlines()]
        ...     diff = difflib.ndiff(stripped(old), stripped(new))
        ...     print '\n'.join(l.strip() for l in diff)

        >>> widget_text = widget.script
        >>> widget_text = widget_text.replace(
        ...     widget.editor_var_name,
        ...     'CKeditor_html').strip()

        >>> print_diff(editor(), widget_text)
        - <textarea cols="60" id="field.html" name="field.html" rows="15" ></textarea>
        <script type="text/javascript" language="JavaScript">
        - var CKeditor_html = new CKEDITOR.replace("field.html",
        ?                                           ------
        + var CKeditor_html = new CKEDITOR.replace("html",
        {
        height: 300,
        + width: 430,
        - customConfig : "/@@/editor_config.js",
        ?             -                        -
        + customConfig: "http://127.0.0.1/@@/editor_config.js"
        ?                 ++++++++++++++++
        }
        );
        </script>

    """


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE |
                   doctest.ELLIPSIS |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    suite = doctest.DocTestSuite(optionflags=optionflags,
                                 setUp=setUp, tearDown=tearDown)
    suite.layer = schoolyear_functional_layer
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool calendar views.
"""
import base64
import datetime

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import adapts, queryMultiAdapter
from zope.i18n import translate
from zope.interface import implements
from zope.publisher.interfaces.browser import IBrowserPublisher
from zope.publisher.interfaces import NotFound
from zope.traversing.api import getParent

from schooltool.app.browser.cal import DailyCalendarView
from schooltool.app.browser.cal import WeeklyCalendarView
from schooltool.app.browser.cal import MonthlyCalendarView
from schooltool.app.browser.cal import YearlyCalendarView
from schooltool.app.browser.cal import CalendarViewBase
from schooltool.app.browser.cal import month_names
from schooltool.app.browser.cal import short_day_of_week_names
from schooltool.app.browser.cal import day_of_week_names
from schooltool.calendar.interfaces import ICalendar
from schooltool.calendar.utils import weeknum_bounds, prev_month, next_month
from schooltool.common.inlinept import InheritTemplate
from schooltool.skin import flourish

from schooltool.common import SchoolToolMessage as _


class CalendarTraverser(object):
    """A smart calendar traverser that can handle dates in the URL."""

    adapts(ICalendar)
    implements(IBrowserPublisher)

    queryMultiAdapter = staticmethod(queryMultiAdapter)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def browserDefault(self, request):
        return self.context, ('daily.html', )

    def publishTraverse(self, request, name):
        view_name = self.getHTMLViewByDate(request, name)
        if not view_name:
            view_name = self.getPDFViewByDate(request, name)
        if view_name:
            return self.queryMultiAdapter((self.context, request),
                                          name=view_name)

        view = queryMultiAdapter((self.context, request), name=name)
        if view is not None:
            return view

        try:
            event_id = base64.decodestring(name).decode("utf-8")
        except:
            raise NotFound(self.context, name, request)

        try:
            return self.context.find(event_id)
        except KeyError:
            raise NotFound(self.context, event_id, request)

    def getHTMLViewByDate(self, request, name):
        """Get HTML view name from URL component."""
        return self.getViewByDate(request, name, 'html')

    def getPDFViewByDate(self, request, name):
        """Get PDF view name from URL component."""
        if not name.endswith('.pdf'):
            return None
        name = name[:-4] # strip off the .pdf
        view_name = self.getViewByDate(request, name, 'pdf')
        if view_name == 'yearly.pdf':
            return None # the yearly PDF view is not available
        else:
            return view_name

    def getViewByDate(self, request, name, suffix):
        """Get view name from URL component."""
        parts = name.split('-')

        if len(parts) == 2 and parts[1].startswith('w'): # a week was given
            try:
                year = int(parts[0])
                week = int(parts[1][1:])
            except ValueError:
                return
            request.form['date'] = self.getWeek(year, week).isoformat()
            return 'weekly.%s' % suffix

        # a year, month or day might have been given
        try:
            parts = [int(part) for part in parts]
        except ValueError:
            return
        if not parts:
            return
        parts = tuple(parts)

        if not (1900 < parts[0] < 2100):
            return

        if len(parts) == 1:
            request.form['date'] = "%d-01-01" % parts
            return 'yearly.%s' % suffix
        elif len(parts) == 2:
            request.form['date'] = "%d-%02d-01" % parts
            return 'monthly.%s' % suffix
        elif len(parts) == 3:
            request.form['date'] = "%d-%02d-%02d" % parts
            return 'daily.%s' % suffix

    def getWeek(self, year, week):
        """Get the start of a week by week number.

        The Monday of the given week is returned as a datetime.date.

            >>> traverser = CalendarTraverser(None, None)
            >>> traverser.getWeek(2002, 11)
            datetime.date(2002, 3, 11)
            >>> traverser.getWeek(2005, 1)
            datetime.date(2005, 1, 3)
            >>> traverser.getWeek(2005, 52)
            datetime.date(2005, 12, 26)

        """
        return weeknum_bounds(year, week)[0]


class FlourishCalendarView(flourish.page.WideContainerPage):
    pass


def year_week(date):
    """Return year and week for given date."""
    return date.isocalendar()[:-1]


class FlourishDailyCalendarView(FlourishCalendarView,
                                DailyCalendarView):
    update = DailyCalendarView.update
    current_title = _("Today")

    @property
    def subtitle(self):
        if self.cursor == self.today:
            return _("Today")
        if self.cursor == self.today + datetime.timedelta(days=1):
            return _("Tomorrow")
        if self.cursor == self.today - datetime.timedelta(days=1):
            return _("Yesterday")
        if (year_week(self.cursor) == year_week(self.today) and
            self.cursor > self.today):
            return day_of_week_names[self.cursor.weekday()]
        return None

    @property
    def date_title(self):
        return DailyCalendarView.title(self)


class FlourishWeeklyCalendarView(FlourishCalendarView,
                                 WeeklyCalendarView):
    update = WeeklyCalendarView.update
    current_title = _("Today")

    @property
    def subtitle(self):
        cursor_yw = year_week(self.cursor)
        if cursor_yw == year_week(self.today):
            return _("This Week")
        if cursor_yw == year_week(self.today - datetime.timedelta(weeks=1)):
            return _("Last Week")
        if cursor_yw == year_week(self.today + datetime.timedelta(weeks=1)):
            return _("Next Week")
        return None

    @property
    def date_title(self):
        return WeeklyCalendarView.title(self)

    def getCurrentWeek(self):
        result = []
        for n, day in enumerate(WeeklyCalendarView.getCurrentWeek(self)):
            day.css_class = ' '.join(filter(None,
               ['day-title', day.css_class, 'first' if n==0 else '']))
            result.append(day)
        return result


class FlourishMonthlyCalendarView(FlourishCalendarView,
                                  MonthlyCalendarView):
    update = MonthlyCalendarView.update
    current_title = _("Today")

    @property
    def subtitle(self):
        return month_names[self.cursor.month]

    @property
    def date_title(self):
        return MonthlyCalendarView.title(self)


class FlourishYearlyCalendarView(FlourishCalendarView,
                                 YearlyCalendarView):
    update = YearlyCalendarView.update
    current_title = _("Today")

    def subtitle(self):
        return unicode(self.cursor.year)

    @property
    def date_title(self):
        return YearlyCalendarView.title(self)


class CalendarJumpTo(flourish.page.Refine):

    title = _('Jump To')
    body_template = ViewPageTemplateFile('templates/calendar_jump_to.pt')

    def getJumpToYears(self):
        """Return jump targets for five years centered on the current year."""
        this_year = self.view.today.year
        return [{'selected': year == this_year,
                 'label': year,
                 'href': self.view.calURL('yearly',
                                          datetime.date(year, 1, 1))}
                for year in range(this_year - 2, this_year + 3)]

    def getJumpToMonths(self):
        """Return a list of months for the drop down in the jump portlet."""
        year = self.view.cursor.year
        return [{'label': v,
                 'href': self.view.calURL('monthly',
                                          datetime.date(year, k, 1))}
                for k, v in month_names.items()]


class CalendarMonthViewlet(flourish.page.Refine):

    body_template = ViewPageTemplateFile('templates/calendar_month_viewlet.pt')

    @property
    def cursor(self):
        return self.view.cursor

    @property
    def month_title(self):
        return month_names[self.cursor.month]

    @property
    def cal_url(self):
        return self.view.calURL('monthly', self.cursor)

    @property
    def rows(self):
        result = []
        cursor = self.cursor
        month = self.view.getMonth(cursor)
        for week in month:
            result.append(self.view.renderRow(week, cursor.month))
        return result

    def weekdays(self):
        result = []
        cursor = self.cursor
        month = self.view.getMonth(cursor)
        week = month[0]
        for day in week:
            name = short_day_of_week_names[day.date.weekday()]
            result.append(translate(name, context=self.request)[:1])
        return result


class CalendarPrevMonthViewlet(CalendarMonthViewlet):

    @Lazy
    def cursor(self):
        return prev_month(self.view.cursor)


class CalendarNextMonthViewlet(CalendarMonthViewlet):

    @Lazy
    def cursor(self):
        return next_month(self.view.cursor)


class CalendarTomorrowEvents(flourish.page.Refine):

    body_template = ViewPageTemplateFile('templates/calendar_tomorrow_events.pt')

    title = _("Tomorrow's Events")

    @property
    def cursor(self):
        today = self.view.today
        tomorrow = today + today.resolution
        return tomorrow

    def get_start_time(self, event, timezone):
        if event.allday:
            # We could make it clearer that this is an all day event
            return ""
        else:
            return event.dtstart.astimezone(timezone).strftime('%H:%M')

    @Lazy
    def events(self):
        cursor = self.cursor
        all_events = self.view.dayEvents(cursor)
        timezone = self.view.timezone
        result = [{'event': e, 'time': self.get_start_time(e, timezone)}
                  for e in all_events
                  if e.dtstart.date() == cursor]
        return result

    def render(self, *args, **kw):
        if not self.events:
            return ''
        return flourish.page.Refine.render(self, *args, **kw)


class CalendarTertiaryNavigation(flourish.page.Content,
                                 flourish.page.TertiaryNavigationManager):
    template = ViewPageTemplateFile('templates/calendar_tertiary_nav.pt')

    mode_types = (
        ('daily', _('Daily')),
        ('weekly', _('Weekly')),
        ('monthly', _('Monthly')),
        ('yearly', _('Yearly')),
        )

    @property
    def nav_today_class(self):
        is_today = self.view.inCurrentPeriod(self.view.today)
        cls = 'calendar-nav'
        if is_today:
            cls += ' active'
        return cls

    @property
    def modes(self):
        result = []
        for mode, title in self.mode_types:
            cls = "calendar-type"
            if mode == self.view.cal_type:
                cls += " active"
            result.append({
                    'title': title,
                    'url': self.view.calURL(mode),
                    'class': cls,
                    })
        return result


class CalendarAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for adding links in calendar views."""


class AddEventLink(flourish.page.LinkViewlet):
    @property
    def url(self):
        return "add.html?field.start_date=%s" % self.view.cursor.isoformat()


def getParentTitleContent(context, request, view):
    parent = getParent(context)
    if parent is None:
        return None
    providers = queryMultiAdapter(
        (parent, request, view),
        flourish.interfaces.IContentProviders)
    if providers is None:
        return None
    content = providers.get("title")
    return content



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.relationship.annotatable
"""

import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements
from zope.annotation.interfaces import IAttributeAnnotatable


def doctest_getRelationshipLinks():
    r"""Test for schooltool.relationship.annotatable.getRelationshipLinks.

    We need to set up Zope 3 annotations

        >>> setup.placelessSetUp()
        >>> setup.setUpAnnotations()

    We need to have an annotatable object

        >>> class SomeAnnotatable(object):
        ...     implements(IAttributeAnnotatable)

        >>> obj = SomeAnnotatable()

    Now we can check that a new LinkSet is created automatically

        >>> from schooltool.relationship.annotatable \
        ...         import getRelationshipLinks
        >>> linkset = getRelationshipLinks(obj)

        >>> from schooltool.relationship.interfaces import IRelationshipLinks
        >>> from zope.interface.verify import verifyObject
        >>> verifyObject(IRelationshipLinks, linkset)
        True

    If you do it more than once, you will get the same link set

        >>> linkset is getRelationshipLinks(obj)
        True

    Cleanup

        >>> setup.placelessTearDown()

    """


def test_suite():
    return unittest.TestSuite([
                doctest.DocTestSuite(),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2012 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
RML tables.
"""
import math
import os

import zope.component
from zope.interface import implements, Interface
from zope.cachedescriptors.property import Lazy

from reportlab.lib import units
from reportlab.pdfbase import pdfmetrics

from schooltool.skin import flourish
from schooltool.table import interfaces
from schooltool.table.column import unindex

from schooltool.common import SchoolToolMessage as _


def minimal_escape_rml(value):
    return value and unicode(value).replace(
        '&', '&amp;').replace(
        '<', '&lt;').replace(
        '>', '&gt;').strip() or ''


def getRMLTable(formatter, request):
    table = zope.component.queryMultiAdapter(
        (formatter.context, request, formatter.view, formatter),
        interfaces.IRMLTable)
    return table


class FlatRMLTable(flourish.content.ContentProvider):
    implements(interfaces.IRMLTable)
    zope.component.adapts(Interface,
                          flourish.interfaces.IFlourishLayer,
                          flourish.interfaces.IPageBase,
                          interfaces.ITableFormatter)

    __name__ = '@@rml'

    formatter = None

    template = flourish.templates.XMLFile('rml/table.pt')

    def __init__(self, context, request, schooltool_formatter, table):
        flourish.content.ContentProvider.__init__(
            self, context, request, schooltool_formatter)
        self.table = table

    @property
    def visible_column_names(self):
        if self.formatter is None:
            return ()
        return [c.name for c in self.formatter.visible_columns]

    def makeFormatter(self):
        if self.table._table_formatter is None:
            # XXX: implicit updates are not a good idea.
            self.table.update()
        formatter = self.table.makeFormatter()
        return formatter

    def update(self):
        self.formatter = self.makeFormatter()

    def getColumns(self):
        visible_names = list(self.visible_column_names)
        # XXX: also add columns we want, remove columns we want to hide
        columns = [self.formatter.columns_by_name[name]
                   for name in visible_names]
        return columns

    def getRMLColumns(self, columns):
        columns = filter(lambda c: c is not None and c.visible, [
            zope.component.queryMultiAdapter(
                (self.context, self.request, self.view, column),
                interfaces.IRMLColumn,
                column.name
                )
            or
            zope.component.queryMultiAdapter(
                (self.context, self.request, self.view, column),
                interfaces.IRMLColumn,
                )
            for column in columns])

        return columns

    def getItems(self):
        # reset batching
        self.formatter.batch_start = self.formatter.batch_size = None
        items = self.formatter.getItems()
        for item in items:
            yield item

    def getRMLId(self, prefix=''):
        parent = self.__parent__
        parents = []
        while (parent is not None and
               flourish.interfaces.IPDFPage.providedBy(parent)):
            name = getattr(parent, '__name__', None)
            parents.append(str(name))
            parent = parent.__parent__
        name_list = ([str(self.__name__)] +
                     parents[:-1] +
                     [prefix])
        return flourish.page.sanitize_id('-'.join(reversed(name_list)))

    def getColumnWidths(self, rml_columns):
        n_cols = len(rml_columns)
        col_width = max(int(100./n_cols), 1)
        result = ' '.join(
            ['%d%%' % col_width] * (n_cols-1) +
            ['%d%%' % max(100-col_width*(n_cols-1), 1)])
        return result

    def getHeaders(self, rml_columns):
        result = [[column.renderHeader(self.formatter)
                   for column in rml_columns]]
        return result

    def render(self):
        columns = self.getColumns()
        rml_columns = self.getRMLColumns(columns)

        if not rml_columns:
            return ''

        items = self.getItems()

        widths_string = self.getColumnWidths(rml_columns)
        headers = self.getHeaders(rml_columns)
        tables = [
            {'rows': [[column.renderCell(item, self.formatter)
                       for column in rml_columns]
                      for item in items]
             }]

        rml = self.template(
            headers=headers, content=tables,
            col_widths=widths_string)
        return rml


class RMLTable(FlatRMLTable):

    def getColumns(self):
        columns = super(RMLTable, self).getColumns()
        group_by_column = self.formatter.group_by_column
        if group_by_column:
            columns = [column
                       for column in columns
                       if column.name != group_by_column]
        return columns

    def render(self):
        columns = self.getColumns()
        rml_columns = self.getRMLColumns(columns)

        if not rml_columns:
            return ''

        items = self.getItems()

        widths_string = self.getColumnWidths(rml_columns)
        headers = self.getHeaders(rml_columns)
        tables = []
        rows = []
        current_group = None
        for item in items:
            group = minimal_escape_rml(self.formatter.getSubGroup(item))
            if group != current_group:
                if rows:
                    tables.append({
                        'headers': current_group and [[current_group]] or [],
                        'rows': rows,
                        })
                current_group = group
                rows = []
            rows.append([column.renderCell(item, self.formatter)
                         for column in rml_columns])

        if rows:
            tables.append({
                    'headers': current_group and [[current_group]] or [],
                    'rows': rows,
                    })

        rml = self.template(
            headers=headers, content=tables,
            col_widths=widths_string)
        return rml


class IndexedRMLTable(RMLTable):
    zope.component.adapts(Interface,
                          flourish.interfaces.IFlourishLayer,
                          flourish.interfaces.IPageBase,
                          interfaces.IIndexedTableFormatter)

    def getColumns(self):
        columns = super(IndexedRMLTable, self).getColumns()
        # Make sure columns can handle indexed objects
        indexed_columns = [interfaces.IIndexedColumn(c) for c in columns]
        return indexed_columns


class RMLTablePart(flourish.report.PDFPart, RMLTable):

    template = flourish.templates.XMLFile('rml/table_part.pt')

    title = None
    table_name = 'table'

    render = RMLTable.render

    def __init__(self, context, request, view, manager):
        flourish.report.PDFPart.__init__(self, context, request, view, manager)

    @Lazy
    def table(self):
        table = self.view.providers[self.table_name]
        return table

    def update(self):
        flourish.report.PDFPart.update(self)
        RMLTable.update(self)

    def getColumns(self):
        table = self.table
        columns = RMLTable.getColumns(self)
        if interfaces.IIndexedTableFormatter.providedBy(table):
            indexed_columns = [interfaces.IIndexedColumn(c) for c in columns]
            return indexed_columns
        return columns


class RMLColumn(object):
    implements(interfaces.IRMLColumn)

    visible = True

    def __init__(self, context, request, view, column):
        self.context = context
        self.request = request
        self.view = view
        self.column = column

    @property
    def name(self):
        return self.column.name

    @property
    def title(self):
        return self.column.title

    def renderHeader(self, formatter):
        header = self.escape(self.column.renderHeader(formatter))
        return header

    def renderCell(self, item, formatter):
        return ''

    escape = lambda self, value: minimal_escape_rml(value)


class HiddenRMLColumn(RMLColumn):
    visible = False


class RMLGetterColumn(RMLColumn):

    def renderCell(self, item, formatter):
        value = self.column.getter(item, formatter)
        return self.escape(value)


class RMLDateColumn(RMLColumn):

    def renderCell(self, item, formatter):
        value = self.column.getter(item, formatter)
        view = zope.component.queryMultiAdapter(
            (value, formatter.request),
            name='mediumDate',
            default=lambda: '')
        date = view()
        return date


class RMLIndexedColumn(RMLGetterColumn):

    def renderCell(self, indexed_item, formatter):
        item = unindex(indexed_item)
        cell = super(RMLIndexedColumn, self).renderCell(item, formatter)
        return cell


class Config(dict):

    def __init__(self, default=None, **kw):
        dict.__init__(self)
        if default is not None:
            self.update(default)
        self.update(kw)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if not hasattr(instance, '_schooltool_table_pdf_Config'):
            setattr(instance, '_schooltool_table_pdf_Config', Config())
            instance._schooltool_table_pdf_Config.update(self)
        return instance._schooltool_table_pdf_Config

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def __repr__(self):
        return '<%s (%s)>' % (self.__class__.__name__, dict.__repr__(self))


class GridCell(object):

    style = Config(
        para_class=None,
        )

    def __init__(self, text, item=None, name=None, no_data=False, **kw):
        self.__name__ = name if name is not None else text
        self.item = item
        self.text = text
        self.no_data = no_data
        self.style.update(kw)

    def __call__(self):
        if self.style.para_class:
            return '<para style="%s">%s</para>' % (
                self.style.para_class, self.text)
        return self.text


class GridColumn(GridCell):

    style = Config(
        GridCell.style,
        font_size=8,
        )

    def __cmp__(self, other):
        return cmp((self.__name__, self.item), (other.__name__, other.item))


class GridRow(GridCell):

    style = Config(
        GridCell.style,
        para_class='grade.table.row-title',
        )

    def __cmp__(self, other):
        return cmp((self.__name__, self.item), (other.__name__, other.item))


def random_id(numbers=4):
    return 'id-'+os.urandom(numbers).encode('hex')


class Grid(object):

    config = Config(
        table_style_name = "grade.table.grades",
        title_column_width = 5 * units.cm,
        header_font = 'Ubuntu_Regular',
        header_font_size = 12,
        column_padding = 4,
        )

    rows = None
    columns = None
    getter = None

    template = flourish.templates.XMLFile('rml/grid.pt')

    def __init__(self, rows, columns, data_getter, table_width, style_id=None,
                 config=None, **kw):
        if config is not None:
            self.config.update(config)
        self.config.update(kw)

        if style_id is None:
            self.style_id = self.random_id()
        else:
            self.style_id = style_id
        self.rows = rows
        self.columns = columns
        self.table_width = table_width
        if hasattr(data_getter, 'get'):
            self._getter_data = data_getter
            self.getter = lambda row, column: data_getter.get((row, column))
        else:
            self.getter = data_getter

    random_id = lambda self: random_id()

    def getCell(self, row, column):
        cell = self.getter(row, column)
        if cell is None:
            return GridCell(u'', item=None)
        if isinstance(cell, GridCell):
            return cell
        return GridCell(unicode(cell), item=cell)

    def getMaxTextSize(self, columns, default_font, default_font_size):
        if not columns:
            return 0, 0
        max_text_length = max([
            pdfmetrics.stringWidth(
                    unicode(column.text),
                    column.style.get('font_name', default_font),
                    column.style.get('font_size', default_font_size),
                    )
            for column in columns
            ])
        max_text_width = max([column.style.get('font_size', default_font_size)
                              for column in columns])
        return max_text_length, max_text_width

    def getDataColumnWidth(self, width, columns, font_size):
        n_cols = len(columns)
        if n_cols <= 1:
            return width

        ang = math.cos(45./180*math.pi)

        max_column_text_length, _unused = self.getMaxTextSize(
            columns, self.config.header_font, font_size)

        last_font_size = columns[-1].style.get('font_size', font_size)

        slanted_header_horiz_w = (max_column_text_length + last_font_size) * ang
        slanted_header_vert_w = last_font_size * ang
        slanted_header_w = slanted_header_horiz_w - slanted_header_vert_w/2

        data_col_w = min(
            (width - slanted_header_w) / (n_cols - 0.5),
            width / n_cols)
        return data_col_w

    def getColumnStrings(self, columns, col_width):
        result = []
        ang = math.cos(45./180*math.pi)
        x = ang*self.config.title_column_width
        padding = col_width / 2 + self.config.header_font_size * ang / 2
        font_name = None
        font_size = None
        for n, column in enumerate(self.columns):
            pos = x + (n * col_width + padding) * ang
            col_font = column.style.get('font_name', self.config.header_font)
            col_font_size = column.style.get('font_size', self.config.header_font_size)
            changed = (col_font != font_name or col_font_size != font_size)
            result.append({
                    'font_changed': changed,
                    'font_name': col_font,
                    'font_size': col_font_size,
                    'x': pos - 2,
                    'y': -pos + 4,
                    'text': column.text,
                    })
            font_name = col_font
            font_size = col_font_size
        return result

    def updateColumns(self, columns, table_width):
        data_width = table_width - self.config.title_column_width
        self.data_column_width = self.getDataColumnWidth(
            data_width, columns, self.config.header_font_size)
        self.column_strings = self.getColumnStrings(
            columns, self.data_column_width)

        n_cols = len(self.columns)
        spacing = max(0, self.table_width
                         - self.config.title_column_width
                         - self.data_column_width*n_cols)
        widths = ([self.config.title_column_width] +
                  [self.data_column_width] * n_cols +
                  [spacing])
        self.column_widths = ' '.join([str(width) for width in widths])

    def updateHeaderGraphics(self):
        self.header_width = self.table_width

        max_text_length, max_text_width = self.getMaxTextSize(
            self.columns, self.config.header_font, self.config.header_font_size)

        ang = math.cos(45./180*math.pi)
        line_len = max_text_length + max_text_width
        self.header_height = line_len * ang

        left = self.config.title_column_width * ang
        self.header_lines = []
        for n in range(len(self.columns)+1):
            pos = left + n*self.data_column_width*ang
            self.header_lines.append(
                '%d %d %d %d' % (pos, -pos, pos+line_len, -pos))

    def updateData(self):
        self.data = [
            [row] + [self.getCell(row, column)
                     for column in self.columns]
            for row in self.rows
            ]

    def getStyleDirectives(self, cell, x, y):
        directives = {}
        font = dict(
             filter(lambda v: v[1], [
                    ('name', cell.style.get('font_name')),
                    ('size', cell.style.get('font_size')),
                    ('leading', cell.style.get('font_leading')),
                    ]))
        if font:
            directives['blockFont'] = sorted(font.items())

        color = cell.style.get('color')
        if color:
            directives['blockTextColor'] = [('colorName', color)]

        bg_color = cell.style.get('background_color')
        if bg_color:
            directives['blockBackground'] = [('colorName', bg_color)]

        align = cell.style.get('align')
        if align:
            directives['blockAlignment'] = [('value', align)]

        valign = cell.style.get('valign')
        if valign:
            directives['blockValign'] = [('value', valign)]

        span_rows = cell.style.get('span_rows')
        span_cols = cell.style.get('span_cols')
        if span_rows or span_cols:
            span_rows = int(span_rows or 1)
            span_cols = int(span_cols or 1)
            directives['blockSpan'] = [
                ('start', '%d %d' % (x, y)),
                ('stop', '%d %d' % (x+span_cols-1, y+span_rows-1)),
                ]
        padding = cell.style.get('padding')
        if padding is not None:
            if isinstance(padding, (tuple, list)):
                if len(padding) == 2:
                    left, top = padding
                    right = left
                    bottom = top
                else:
                    left, top, right, bottom = padding
            else:
                left = top = bottom = right = float(padding)
            directives['blockLeftPadding'] = [('length', left)]
            directives['blockTopPadding'] = [('length', top)]
            directives['blockRightPadding'] = [('length', right)]
            directives['blockBottomPadding'] = [('length', bottom)]
        return directives

    def updateStyles(self):
        style_map = {}
        for y, row in enumerate(self.data):
            for x, cell in enumerate(row):
                directives = self.getStyleDirectives(cell, x, y)
                for directive, signature in directives.items():
                    if directive not in style_map:
                        style_map[directive] = {}
                    if signature not in style_map[directive]:
                        style_map[directive][signature] = [(x, y)]
                    else:
                        style_map[directive][signature].append((x, y))
        self.styles = {}
        for directive, signatures in sorted(style_map.items()):
            if directive not in self.styles:
                self.styles[directive] = []
            for signature, coordinates in sorted(signatures.items()):
                blocks = makeCoordinateBlocks(coordinates)
                for starts, ends in blocks:
                    style = {'start': '%d %d' % starts,
                             'stop': '%d %d' % ends}
                    style.update(dict(list(signature)))
                    self.styles[directive].append(style)

    def update(self):
        self.updateColumns(self.columns, self.table_width)
        self.updateHeaderGraphics()
        self.updateData()
        self.updateStyles()

    def renderStyles(self):
        result = []
        for directive in self.styles:
            for entries in self.styles[directive]:
                for entry in entries:
                    result.append('<%s %s />' % (
                        directive,
                        ' '.join(['%s="%s"' % (k, v)
                                  for k, v in sorted(entry.items())])
                        ))
        return '\n'.join(result)

    def render(self):
        return self.template()


class AutoFitGrid(Grid):

    config = Config(
        Grid.config,
        header_min_font_size = 12,
        min_column_width = 1.5 * units.cm,
        max_column_width = 1.5 * units.cm,
        continued_font = Grid.config.header_font,
        continued_font_size = 8,
        continued_text = _('Continued ...'),
        )

    remaining_columns = None

    def getDataColumnWidth(self, width, columns, font_size):
        data_column_width = Grid.getDataColumnWidth(self, width, columns, font_size)
        if self.config.max_column_width is None:
            return data_column_width
        min_headers_column_width = self.getMinHeadersWidth(columns, font_size)
        return max(min_headers_column_width,
                   min(data_column_width, self.config.max_column_width)
                )

    def getMinHeadersWidth(self, columns, font_size):
        ang = math.cos(45./180*math.pi)
        _unused, max_text_height = self.getMaxTextSize(
            columns, self.config.header_font, font_size)
        # Min width so that headers would fit
        min_headers_column_width = (max_text_height * ang
                                    + max_text_height * ang
                                    + self.config.column_padding * 2)
        return min_headers_column_width

    def columnsFit(self, columns, table_width, font_size):
        min_headers_column_width = self.getMinHeadersWidth(columns, font_size)
        data_column_width = self.getDataColumnWidth(table_width, columns, font_size)
        if self.config.min_column_width is None:
            return data_column_width >= min_headers_column_width
        # Column must fit headers and honour desired min_column_width
        min_column_width = max(min_headers_column_width,
                               self.config.min_column_width)
        return data_column_width >= min_column_width

    def fitColumns(self, columns, table_width):
        if not columns:
            return

        # Try fitting all columns, reducing the font if necessary
        orig_font_size = self.config.header_font_size
        font_size = self.config.header_font_size
        while font_size >= self.config.header_min_font_size:
            if self.columnsFit(columns, table_width, font_size):
                self.columns = columns
                self.remaining_columns = []
                self.config.header_font_size = font_size
                return
            font_size -= 1
        self.config.header_font_size = orig_font_size

        # Fit as many columns as we can
        continued = [GridColumn(self.config.continued_text,
                                font_name=self.config.continued_font,
                                font_size=self.config.continued_font_size,
                                no_data=True)]
        n = 1
        while n < len(columns):
            if not self.columnsFit(
                columns[:n] + continued, table_width, self.config.header_font_size):
                break
            n += 1
        self.columns = columns[:n] + continued
        self.remaining_columns = columns[n:]

    def updateColumns(self, columns, table_width):
        self.all_columns = columns
        self.remaining_columns = []
        self.fitColumns(columns, table_width)
        super(AutoFitGrid, self).updateColumns(self.columns, table_width)


class GridContent(flourish.content.ContentProvider):

    tables = None

    def getTableWidth(self):
        return self.view.page_size[0] - self.view.margin.left - self.view.margin.right

    def getRMLId(self, prefix=''):
        parent = self.__parent__
        parents = []
        while (parent is not None and
               flourish.interfaces.IPDFPage.providedBy(parent)):
            name = getattr(parent, '__name__', None)
            parents.append(str(name))
            parent = parent.__parent__
        name_list = ([str(self.__name__)] +
                     parents[:-1] +
                     [prefix])
        return flourish.page.sanitize_id('-'.join(reversed(name_list)))

    def updateTables(self):
        id_base = self.getRMLId()
        table_width = self.getTableWidth()
        table = AutoFitGrid(
            self.rows, self.columns, self.grid, table_width,
            style_id=id_base+'-style')
        table.update()
        self.tables = [table]
        while table.remaining_columns:
            other = AutoFitGrid(
                self.rows, table.remaining_columns, self.grid, table_width,
                style_id=id_base+'-style%d' % len(self.tables),
                config=table.config)
            other.update()
            self.tables.append(other)
            table = other

    def update(self):
        super(GridContent, self).update()
        self.updateGrid()
        self.updateTables()

    def render(self):
        return '\n'.join([table.render() for table in self.tables])

    def updateGrid(self):
        # List of GridColumn
        self.columns = []
        # List of GridRow
        self.rows = []
        # data[GridRow, GridColumn] = GridCell / number / string
        self.grid = {}


class GridContentBlock(flourish.report.PDFContentBlock, GridContent):

    template = flourish.templates.XMLFile('rml/grid_part.pt')
    render = template

    update = GridContent.update


class GridPart(flourish.report.PDFPart, GridContentBlock):

    template = flourish.templates.XMLFile('rml/grid_part.pt')
    render = template

    def update(self):
        flourish.report.PDFPart.update(self)
        GridContentBlock.update(self)

    def __init__(self, context, request, view, manager):
        GridContent.__init__(self, context, request, view)
        flourish.report.PDFPart.__init__(self, context, request, view, manager)


def makeCoordinateBlocks(coords):
    if not coords:
        return []
    if len(coords) == 1:
        return [(coords[0], coords[0])]
    coords = sorted(coords, key=lambda c: (c[1], c[0]))
    slices = []

    sx, sy = coords[0]
    ex = sx
    for x, y in coords[1:]:
        if y == sy and x == ex+1:
            ex = x
        else:
            slices.append(((sx, sy), (ex, sy)))
            sx = ex = x
            sy = y
    if coords[-1] != slices[-1][-1]:
        slices.append(((sx, sy), (ex, sy)))

    blocks = []
    while slices:
        slice = slices.pop(0)
        (sx, sy), (ex, ey) = slice
        for other in list(slices):
            if (other[0][0] == sx and
                other[1][0] == ex and
                other[0][1] == ey+1):
                ey = ey + 1
                slices.remove(other)
            elif (ex > sx and
                  other[0][0] == sx and
                  other[1][0] > ex and
                  other[0][1] == ey+1):
                ey = ey + 1
                slices[slices.index(other)] = ((ex+1, other[0][1]),
                                               (other[1][0], other[1][1]))
        blocks.append(((sx, sy), (ex, ey)))
    return blocks



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.generations.evolve30
"""
import unittest
import doctest

from zope.app.testing import setup
from zope.interface import implements
from zope.container.contained import Contained
from zope.container.btree import BTreeContainer
from zope.component import provideHandler

from schooltool.relationship.tests import setUpRelationships
from schooltool.generations.tests import ContextStub
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.person import BasicPerson
from schooltool.group.group import Group
from schooltool.relationship.interfaces import IRelationshipRemovedEvent
from schooltool.relationship.interfaces import IRelationshipAddedEvent
from schooltool.relationship.relationship import relate
from schooltool.app.membership import URIMember, URIGroup, URIMembership
from schooltool.course.interfaces import ICourse


class AppStub(BTreeContainer):
    implements(ISchoolToolApplication)


class CourseStub(Contained):
    implements(ICourse)
    course_id = None

    def __repr__(self):
        return '<CourseStub(__name__="%s", course_id="%s")>' % (
            self.__name__, self.course_id)


def printRelationshipAdded(event):
    if event.rel_type != URIMembership:
        return
    print 'ADD: %15s to   %s' % (
        event[URIMember].title, event[URIGroup].title)


def printRelationshipRemoved(event):
    if event.rel_type != URIMembership:
        return
    print 'REMOVE: %12s from %s' % (
        event[URIMember].title, event[URIGroup].title)


def doctest_evolve30():
    """Evolution to generation 30.

        >>> context = ContextStub()
        >>> context.root_folder['app'] = app = AppStub()

        >>> persons = app['persons'] = BTreeContainer()
        >>> persons['will'] = BasicPerson("will", "William", "Straus")
        >>> persons['vlad'] = BasicPerson("vlad", "Vladimir", "Rubov")
        >>> persons['john'] = BasicPerson("john", "Johny", "John")
        >>> persons['pete'] = BasicPerson("pete", "Petey", "Pete")
        >>> persons['bill'] = BasicPerson("bill", "Billy", "Bill")

        >>> groups = app['groups'] = BTreeContainer()
        >>> groups['guests'] = Group('Guests')
        >>> groups['students'] = Group('Students')

        >>> sections = app['sections'] = BTreeContainer()
        >>> sections['section_1'] = Group('Section ONE')
        >>> sections['section_2'] = Group('Section TWO')

        >>> def addMember(member, group):
        ...     relate(URIMembership,
        ...            (member, URIMember),
        ...            (group, URIGroup))

        >>> addMember(persons['john'], groups['students'])
        >>> addMember(persons['pete'], groups['students'])
        >>> addMember(persons['bill'], groups['students'])

        >>> addMember(persons['will'], groups['guests'])
        >>> addMember(persons['vlad'], groups['guests'])

    Set up section relationships.

    Add one guest to Section ONE, but not the group.

        >>> addMember(persons['vlad'], sections['section_1'])

    Add both guests to Section TWO and the group.

        >>> addMember(groups['guests'], sections['section_2'])
        >>> addMember(persons['vlad'], sections['section_2'])
        >>> addMember(persons['will'], sections['section_2'])

    Add just the students group to Section ONE.

        >>> addMember(groups['students'], sections['section_1'])

    Add two of three students to Section TWO, and the students group.

        >>> addMember(groups['students'], sections['section_2'])
        >>> addMember(persons['bill'], sections['section_2'])
        >>> addMember(persons['pete'], sections['section_2'])

    We will also update course ids, so let's set up some.

        >>> courses = app['courses'] = BTreeContainer()
        >>> courses['c1'] = CourseStub()
        >>> courses['c2'] = CourseStub()
        >>> courses['c3'] = CourseStub()
        >>> courses['c3'].course_id = 'custom'
        >>> list(courses.values())
        [<CourseStub(__name__="c1", course_id="None")>,
         <CourseStub(__name__="c2", course_id="None")>,
         <CourseStub(__name__="c3", course_id="custom")>]

    Let's evolve now.

        >>> provideHandler(printRelationshipAdded,
        ...                [IRelationshipAddedEvent])
        >>> provideHandler(printRelationshipRemoved,
        ...                [IRelationshipRemovedEvent])

        >>> from schooltool.generations.evolve30 import evolve
        >>> evolve(context)
        REMOVE:       Guests from Section TWO
        <BLANKLINE>
        ADD:     Johny John to   Section ONE
        ADD:     Petey Pete to   Section ONE
        ADD:     Billy Bill to   Section ONE
        REMOVE:     Students from Section ONE
        <BLANKLINE>
        ADD:     Johny John to   Section TWO
        REMOVE:     Students from Section TWO

    Course ids are also updated.

        >>> list(courses.values())
        [<CourseStub(__name__="c1", course_id="c1")>,
         <CourseStub(__name__="c2", course_id="c2")>,
         <CourseStub(__name__="c3", course_id="custom")>]

    """


def setUp(test):
    setup.placelessSetUp()
    setup.setUpAnnotations()
    setUpRelationships()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = (doctest.ELLIPSIS |
                   doctest.NORMALIZE_WHITESPACE |
                   doctest.REPORT_ONLY_FIRST_FAILURE)
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')
